"""
obsah constraints module

implements validation for constraints used in obsah CLIs
"""
import argparse

def validate_constraints(metadata: dict, args: argparse.Namespace):
    """
    validate arguments passed in on the CLI against constraints from a playbook
    """
    constraints: dict = metadata.get('constraints', {})
    variables: list = metadata.get('variables', [])

    def variable_to_parameter(name):
        for variable in variables:
            if variable.name == name:
                return variable.parameter
        return name

    errors = []
    for constraint in constraints.get('required_together', []):
        present_args = [arg in args for arg in constraint]
        if not all(present_args) and any(present_args):
            errors.append(f"{[variable_to_parameter(x) for x in constraint]} are required together")
    for constraint in constraints.get('required_one_of', []):
        present_args = [arg in args for arg in constraint]
        if not any(present_args):
            errors.append(f"one of {[variable_to_parameter(x) for x in constraint]} is required")
    for constraint in constraints.get('required_if', []):
        argument_name, argument_value, required_arguments = constraint
        if argument_name in args and getattr(args, argument_name) == argument_value:
            if not all(arg in args for arg in required_arguments):
                required = [variable_to_parameter(x) for x in required_arguments]
                errors.append(f"{required} are required because {variable_to_parameter(argument_name)} is {argument_value}")
    for constraint in constraints.get('forbidden_if', []):
        argument_name, argument_value, forbidden_arguments = constraint
        if argument_name in args and getattr(args, argument_name) == argument_value:
            if any(arg in args for arg in forbidden_arguments):
                forbidden = [variable_to_parameter(x) for x in forbidden_arguments]
                errors.append(f"{forbidden} are forbidden because {variable_to_parameter(argument_name)} is {argument_value}")
    for constraint in constraints.get('mutually_exclusive', []):
        present_args = [True for arg in constraint if arg in args]
        if len(present_args) > 1:
            errors.append(f"{constraint} are mutually exclusive")
    return errors



"""
obsah data types to be used to validate user input
"""

import argparse
import os
import pathlib
import re

REGISTRY: list[type['BaseType']] = []


class BaseType:
    """
    Base type class
    """

    def __init_subclass__(cls, **kwargs):
        if not cls.__name__.endswith('Type'):
            REGISTRY.append(cls)

    @property
    def name(self):
        """
        the name of the type
        this is either the class name of, if set, the `type_name` property
        """
        if hasattr(self, 'type_name'):
            return self.type_name
        return self.__class__.__name__

    def validate(self, string):
        """
        validate user input
        """
        return string


class Boolean(BaseType):
    """
    Boolean type, accepting true/false and 1/0 only
    """
    def validate(self, string):
        if string.lower() in ['true', '1']:
            return True
        if string.lower() in ['false', '0']:
            return False
        raise ValueError


class AbsolutePath(BaseType):
    """
    Path type, accepting anything that looks like an absolute path
    """
    def validate(self, string):
        path = pathlib.PurePath(string)
        if path.is_absolute():
            return path.as_posix()
        raise ValueError


class File(BaseType):
    """
    File type, accepting any existing file
    """
    def validate(self, string):
        if os.path.isfile(string):
            return string
        raise ValueError


class Port(BaseType):
    """
    TCP/UDP Port type, accepting an integer between 0 and 65535
    """
    def validate(self, string):
        if 0 < int(string) < 65535:
            return int(string)
        raise ValueError


class RegexType(BaseType):
    """
    Validates a string against a regular expression
    """
    REGEX = r'.*'

    def validate(self, string):
        if re.match(self.REGEX, string):
            return string
        raise ValueError


class FQDN(RegexType):
    """
    A FQDN
    """
    REGEX = r'\A(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])\Z'


class HTTPUrl(RegexType):
    """
    A HTTP or HTTPS URL
    """
    REGEX = r'(?i:\Ahttps?:\/\/.*\Z)'


def register_types(parser: argparse.ArgumentParser):
    """
    register all known data types as types usable by argparse
    """
    for t in REGISTRY:
        tobj = t()
        parser.register('type', tobj.name, tobj.validate)



#!/usr/bin/env python
# PYTHON_ARGCOMPLETE_OK

"""
Obsah is a wrapper around Ansible playbooks. They are exposed as a command line application.
"""

from __future__ import print_function

import argparse
import glob
import json
import os
from collections import namedtuple
from functools import total_ordering
from importlib import resources

import yaml

from . import data_types
from .constraints import validate_constraints

try:
    import argcomplete
except ImportError:
    argcomplete = None


# Need for PlaybookCLI to set the verbosity
display = None  # pylint: disable=C0103


Variable = namedtuple('Variable', ['name', 'parameter', 'help_text', 'action', 'type', 'choices'])


@total_ordering
class Playbook(object):
    """
    An abstraction over an Ansible playbook
    """
    def __init__(self, path, application_config):
        self.path = path
        self.application_config = application_config
        directory = os.path.dirname(self.path)
        self.name = os.path.basename(directory)
        self._metadata_path = os.path.join(directory, self.application_config.metadata_name())
        self._metadata = None


    def _load_metadata_file(self, path):
        """
        Read JSON from a file
        """
        try:
            with open(path) as obsah_metadata:
                data = yaml.safe_load(obsah_metadata)
        except FileNotFoundError:
            data = {}
        return data


    @property
    def metadata(self):
        """
        Read metadata about the playbook

        The metadata can contain a global help text as well as variables that should be exposed as
        command line parameters.

        This data is lazily loaded and cached.
        """
        if not self._metadata:
            data = self._load_metadata_file(self._metadata_path)
            for include in data.get('include', []):
                include_path = os.path.join(self.application_config.playbooks_path(), include, self.application_config.metadata_name())
                include_data = self._load_metadata_file(include_path)
                data['variables'] = data.get('variables', {}) | include_data.get('variables', {})
                data['constraints'] = data.get('constraints', {}) | include_data.get('constraints', {})

            self._metadata = {
                'help': data.get('help'),
                'variables': sorted(self._parse_parameters(data.get('variables', {}))),
                'constraints': data.get('constraints', {}),
            }

        return self._metadata

    @property
    def takes_target_parameter(self):
        """
        Whether this playbook takes a target argument.

        This is determined by a hosts: targets inside the playbook
        """
        with open(self.path) as playbook_file:
            plays = yaml.safe_load(playbook_file.read())

        target_names = set(self.application_config.target_names())

        return any(len(target_names.intersection(play.get('hosts', []))) > 0 for play in plays)

    @property
    def playbook_variables(self):
        """
        The playbook variables that should be exposed to the user

        This is extracted from the metadata.
        """
        return self.metadata['variables']

    @property
    def help_text(self):
        """
        The help text if available. This is the first line from the help in the metadata.
        """
        return self.metadata['help'].split('\n', 1)[0] if self.metadata['help'] else None

    @property
    def description(self):
        """
        The full help text if available. This is extracted from the metadata.
        """
        return self.metadata['help']

    def _parse_parameters(self, variables):
        """
        Parse parameters from the metadata.

        Automatically determines the parameter if not specified. This is done by looking at the
        variable and de-namespacing if it's namespaced. Also replaces underscores with dashes. This
        means that for the playbook changelog we expose changelog_message as --message but
        other_option as --other-option.
        """
        namespace = '{}_'.format(self.name)

        for name, options in variables.items():
            try:
                parameter = options['parameter']
            except KeyError:
                parameter = '--{}'.format(name.removeprefix(namespace).replace('_', '-'))

            yield Variable(name, parameter, options.get('help'), options.get('action'), options.get('type'), options.get('choices'))

    @property
    def __doc__(self):
        return self.description

    def __str__(self):
        return self.name

    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.path)

    def __eq__(self, other):
        if hasattr(other, 'path'):
            return self.path == other.path
        return NotImplemented

    def __lt__(self, other):
        if hasattr(other, 'name'):
            return self.name.__lt__(other.name)
        return NotImplemented


class ApplicationConfig(object):
    """
    A class describing the where to find various files
    """

    @staticmethod
    def name():
        """
        Return the name as shown to the user in the ArgumentParser
        """
        return 'obsah'

    @staticmethod
    def target_name():
        """
        Return the name of the target in the playbook if the playbook takes a parameter.
        """
        return 'packages'

    @classmethod
    def target_names(cls):
        """
        Return the names of the targets in the playbook if the playbook takes parameters.
        """
        return [cls.target_name()]

    @staticmethod
    def metadata_name():
        """
        Return the name of the metadata file.
        """
        return 'metadata.obsah.yaml'

    @staticmethod
    def data_path():
        """
        Return the data path. Houses playbooks and configs.
        """
        path = os.environ.get('OBSAH_DATA')
        if path is None:
            with resources.path(__name__, 'data') as fspath:
                path = fspath.as_posix()

        return path

    @staticmethod
    def inventory_path():
        """
        Return the inventory path
        """
        return os.environ.get('OBSAH_INVENTORY', os.path.join(os.getcwd(), 'package_manifest.yaml'))

    @classmethod
    def playbooks_path(cls):
        """
        Return the default playbooks path
        """
        return os.environ.get('OBSAH_PLAYBOOKS', os.path.join(cls.data_path(), 'playbooks'))

    @classmethod
    def ansible_config_path(cls):
        """
        Return the ansible.cfg path
        """
        return os.environ.get('OBSAH_ANSIBLE_CFG', os.path.join(cls.data_path(), 'ansible.cfg'))

    @classmethod
    def playbooks(cls):
        """
        Return all playbooks in the playbook path.
        """
        paths = glob.glob(os.path.join(cls.playbooks_path(), '*', '*.yaml'))
        return sorted(Playbook(playbook_path, cls) for playbook_path in paths if
                      os.path.basename(playbook_path) != cls.metadata_name())

    @staticmethod
    def allow_extra_vars():
        """
        Whether to allow --extra-vars parameter to be automatically added
        """
        return True


class ObsahArgumentParser(argparse.ArgumentParser):
    def exit(self, status=0, message=None):
        if message is not None and not message.endswith("\n"):
            message += "\n"
        super().exit(status, message)

    def format_usage(self):
        return super().format_help()


def find_targets(inventory_path):
    """
    Find all targets in the given inventory
    """
    targets = None
    if os.path.exists(inventory_path):
        from ansible.inventory.manager import InventoryManager # pylint: disable=all
        from ansible.parsing.dataloader import DataLoader # pylint: disable=all
        ansible_loader = DataLoader()
        ansible_inventory = InventoryManager(loader=ansible_loader,
                                             sources=inventory_path)
        targets = list(ansible_inventory.hosts.keys())
        targets.extend(ansible_inventory.groups.keys())
    return targets


def obsah_argument_parser(application_config=ApplicationConfig, playbooks=None, targets=None):
    """
    Construct an argument parser with the given actions and target choices.
    """
    if playbooks is None:
        playbooks = application_config.playbooks()

    if targets is None:
        targets = []

    parser = ObsahArgumentParser(application_config.name())

    parser.obsah_arguments = []

    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument("-v", "--verbose",
                               action="count",
                               dest="verbose",
                               help="verbose output")

    if application_config.allow_extra_vars():
        advanced = parent_parser.add_argument_group('advanced arguments')
        advanced.add_argument('-e', '--extra-vars',
                              dest="extra_vars",
                              action="append",
                              default=[],
                              help="""set additional variables as key=value or
                              YAML/JSON, if filename prepend with @""")

    subparsers = parser.add_subparsers(dest='action', metavar='action',
                                       required=True,
                                       help="""which action to execute""")

    for playbook in playbooks:
        subparser = subparsers.add_parser(playbook.name, parents=[parent_parser],
                                          help=playbook.help_text,
                                          description=playbook.description,
                                          formatter_class=argparse.RawDescriptionHelpFormatter)
        data_types.register_types(subparser)
        subparser.set_defaults(playbook=playbook)

        if playbook.takes_target_parameter:
            subparser.add_argument('target',
                                   metavar='target',
                                   choices=targets,
                                   nargs='+',
                                   help="the target to execute the action against")

        for variable in playbook.playbook_variables:
            argument_args = {'help': variable.help_text, 'action': variable.action,
                             'default': argparse.SUPPRESS}
            if variable.type is not None:
                argument_args['type'] = variable.type
            if variable.choices is not None:
                argument_args['choices'] = variable.choices
            if variable.parameter.startswith('--'):
                argument_args['dest'] = variable.name
            subparser.add_argument(variable.parameter, **argument_args)
            parser.obsah_arguments.append(variable.name)

    if argcomplete:
        argcomplete.autocomplete(parser)

    return parser


def generate_ansible_args(inventory_path, args, obsah_arguments):
    """
    Generate the arguments to run ansible based on the parsed command line arguments
    """
    ansible_args = [args.playbook.path, '--inventory', inventory_path]
    if hasattr(args, 'target'):
        limit = ':'.join(args.target)
        ansible_args.extend(['--limit', limit])
    if args.verbose:
        ansible_args.append("-%s" % str("v" * args.verbose))
    for extra_var in getattr(args, 'extra_vars', []):
        ansible_args.extend(["-e", extra_var])

    variables = {arg: getattr(args, arg) for arg in obsah_arguments if hasattr(args, arg)}
    if variables:
        ansible_args.extend(["-e", json.dumps(variables, sort_keys=True)])

    return ansible_args


def main(cliargs=None, application_config=ApplicationConfig):  # pylint: disable=R0914
    """
    Main command
    """
    cfg_path = application_config.ansible_config_path()

    if os.path.exists(cfg_path):
        os.environ["ANSIBLE_CONFIG"] = cfg_path

    # this needs to be global, as otherwise PlaybookCLI fails
    # to set the verbosity correctly
    from ansible.utils.display import Display # pylint: disable=all
    global display  # pylint: disable=C0103,W0603
    display = Display()

    inventory_path = application_config.inventory_path()

    targets = find_targets(inventory_path)

    parser = obsah_argument_parser(application_config, targets=targets)

    args = parser.parse_args(cliargs)

    if errors := validate_constraints(args.playbook.metadata, args):
        parser.exit(1, "\n".join(errors))

    if args.playbook.takes_target_parameter and not os.path.exists(inventory_path):
        parser.exit(1, "Could not find your inventory at {}".format(inventory_path))

    from ansible.cli.playbook import PlaybookCLI # pylint: disable=all

    ansible_args = generate_ansible_args(inventory_path, args, parser.obsah_arguments)
    ansible_playbook = (["ansible-playbook"] + ansible_args)

    if args.verbose:
        print("ANSIBLE_CONFIG={}".format(os.environ["ANSIBLE_CONFIG"]), ' '.join(ansible_playbook))

    cli = PlaybookCLI(ansible_playbook)
    cli.parse()
    exit_code = cli.run()
    parser.exit(exit_code)


if __name__ == '__main__':
    main()



import argparse

import pytest

from obsah.constraints import validate_constraints


@pytest.mark.parametrize("constraints,args,errors", [
    ([['a', 'b']], {'a': 1, 'b': 1}, []),
    ([['a', 'b']], {'a': 1}, ["['a', 'b'] are required together"]),
    ([['a', 'b']], {}, []),
])
def test_required_together(constraints, args, errors):
    metadata = {'constraints': {'required_together': constraints}}
    found_errors = validate_constraints(metadata, argparse.Namespace(**args))
    assert errors == found_errors


@pytest.mark.parametrize("constraints,args,errors", [
    ([['a', 'b']], {'a': 1, 'b': 1}, []),
    ([['a', 'b']], {'a': 1}, []),
    ([['a', 'b']], {}, ["one of ['a', 'b'] is required"]),
])
def test_required_one_of(constraints, args, errors):
    metadata = {'constraints': {'required_one_of': constraints}}
    found_errors = validate_constraints(metadata, argparse.Namespace(**args))
    assert errors == found_errors


@pytest.mark.parametrize("constraints,args,errors", [
    ([['a', 1, ['b']]], {'a': 1, 'b': 1}, []),
    ([['a', 1, ['b']]], {'a': 1}, ["['b'] are required because a is 1"]),
    ([['a', 1, ['b']]], {'a': 2}, []),
])
def test_required_if(constraints, args, errors):
    metadata = {'constraints': {'required_if': constraints}}
    found_errors = validate_constraints(metadata, argparse.Namespace(**args))
    assert errors == found_errors


@pytest.mark.parametrize("constraints,args,errors", [
    ([['a', 1, ['b']]], {'a': 1, 'b': 1}, ["['b'] are forbidden because a is 1"]),
    ([['a', 1, ['b']]], {'a': 1}, []),
    ([['a', 1, ['b']]], {'a': 2, 'b': 1}, []),
])
def test_forbidden_if(constraints, args, errors):
    metadata = {'constraints': {'forbidden_if': constraints}}
    found_errors = validate_constraints(metadata, argparse.Namespace(**args))
    assert errors == found_errors


@pytest.mark.parametrize("constraints,args,errors", [
    ([['a', 'b']], {'a': 1, 'b': 1}, ["['a', 'b'] are mutually exclusive"]),
    ([['a', 'b']], {'a': 1}, []),
    ([['a', 'b']], {}, []),
])
def test_mutually_exclusive(constraints, args, errors):
    metadata = {'constraints': {'mutually_exclusive': constraints}}
    found_errors = validate_constraints(metadata, argparse.Namespace(**args))
    assert errors == found_errors



import sys

import pytest
import obsah


@pytest.fixture
def playbooks_path(fixture_dir):
    return fixture_dir / 'playbooks'


@pytest.fixture
def application_config(playbooks_path):
    class MockApplicationConfig(obsah.ApplicationConfig):
        @staticmethod
        def playbooks_path():
            return playbooks_path.strpath

        @staticmethod
        def target_name():
            return 'packages'

    return MockApplicationConfig


@pytest.fixture
def application_config_multi_target(playbooks_path):
    class MockApplicationConfig(obsah.ApplicationConfig):
        @staticmethod
        def playbooks_path():
            return playbooks_path.strpath

        @staticmethod
        def target_names():
            return ['packages', 'repos']

    return MockApplicationConfig


@pytest.fixture
def parser(application_config, targets=['testpackage']):
    return obsah.obsah_argument_parser(application_config, targets=targets)


def test_find_no_targets(fixture_dir):
    targets = obsah.find_targets((fixture_dir / 'nope.yaml').strpath)
    assert targets is None


def test_find_targets(fixture_dir):
    targets = obsah.find_targets((fixture_dir / 'inventory.yaml').strpath)
    assert targets
    assert 'testpackage' in targets
    assert 'all' in targets
    assert 'repos' in targets


def test_playbook_constructor(application_config, playbooks_path):
    path = (playbooks_path / 'setup' / 'setup.yaml').strpath
    playbook = obsah.Playbook(path, application_config)
    assert playbook.path == path
    assert playbook.name == 'setup'


@pytest.mark.parametrize('playbook,expected', [
    ('setup', False),
    ('positional', False),
    ('dummy', True),
    ('multiple_plays', True),
    ('repoclosure', True),
])
def test_playbook_takes_target_parameter(application_config, playbooks_path, playbook, expected):
    path = (playbooks_path / playbook / '{}.yaml'.format(playbook)).strpath
    assert obsah.Playbook(path, application_config).takes_target_parameter == expected

@pytest.mark.parametrize('playbook,expected', [
    ('setup', False),
    ('positional', False),
    ('dummy', True),
    ('multiple_plays', True),
    ('repoclosure', True),
])
def test_playbook_takes_multi_target_parameter(application_config_multi_target, playbooks_path, playbook, expected):
    path = (playbooks_path / playbook / '{}.yaml'.format(playbook)).strpath
    assert obsah.Playbook(path, application_config_multi_target).takes_target_parameter == expected


def test_parser_no_arguments(parser):
    with pytest.raises(SystemExit):
        parser.parse_args([])


@pytest.mark.parametrize('cliargs,expected', [
    (['setup'],
     []),
    (['positional', 'tag1'],
     ['-e', '{"tag": "tag1"}']),
    (['dummy', 'testpackage'],
     ['--limit', 'testpackage']),
    (['dummy', 'testpackage', '--verbose'],
     ['--limit', 'testpackage', '-v']),
    (['dummy', 'testpackage', '-vvvv'],
     ['--limit', 'testpackage', '-vvvv']),
    (['dummy', 'testpackage', '-e', 'v1=1', '-e', 'v2=2'],
     ['--limit', 'testpackage', '-e', 'v1=1', '-e', 'v2=2']),
    (['dummy', 'testpackage', '--automatic', 'foo'],
     ['--limit', 'testpackage', '-e', '{"automatic": "foo"}']),
    (['dummy', 'testpackage', '--explicit', 'foo'],
     ['--limit', 'testpackage', '-e', '{"mapped": "foo"}']),
    (['dummy', 'testpackage', '--store-true'],
     ['--limit', 'testpackage', '-e', '{"store_true": true}']),
    (['dummy', 'testpackage', '--store-false'],
     ['--limit', 'testpackage', '-e', '{"store_false": false}']),
    (['dummy', 'testpackage', '--automatic', 'foo', '--explicit', 'bar'],
     ['--limit', 'testpackage', '-e', '{"automatic": "foo", "mapped": "bar"}']),
    (['dummy', 'testpackage', '--my-list', 'foo', '--my-list', 'bar'],
     ['--limit', 'testpackage', '-e', '{"mapped_list": ["foo", "bar"]}']),
])
def test_generate_ansible_args(playbooks_path, parser, cliargs, expected):
    action = cliargs[0]
    base_expected = [(playbooks_path / action / '{}.yaml'.format(action)).strpath,
                     '--inventory', 'inventory.yml']

    args = parser.parse_args(cliargs)
    ansible_args = obsah.generate_ansible_args('inventory.yml', args, parser.obsah_arguments)
    assert ansible_args == base_expected + expected


def test_obsah_argument_parser_help(fixture_dir, parser):
    # https://github.com/python/cpython/commit/7cc773ba3d07d4a2e6cd39063fd1954abd6ae8f1
    if sys.version_info >= (3, 12, 7):
        path = fixture_dir / 'help-3.12.txt'
    else:
        path = fixture_dir / 'help.txt'
    expected = path.read()
    if sys.version_info >= (3, 10, 0):
        expected = expected.replace('optional arguments:', 'options:')
    assert expected == parser.format_help()



"""
Setup file for obsah
"""

# To use a consistent encoding
import codecs
import os

# Always prefer setuptools over distutils
from setuptools import setup, find_packages


def get_long_description():
    """
    Get the long description from the README file
    """
    here = os.path.abspath(os.path.dirname(__file__))

    with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as readme:
        return readme.read()


def find_package_data(package, data_dir):
    """
    Find all the package data
    """
    package_data = []
    oldcwd = os.getcwd()
    os.chdir(package)
    for dirpath, _, filenames in os.walk(data_dir):
        files = [os.path.join(dirpath, filename) for filename in filenames]
        package_data.extend(files)
    os.chdir(oldcwd)
    return package_data


setup(
    name='obsah',
    version='1.1.0',
    license='GPL-2.0-only',
    description='packaging wrapper using ansible',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/theforeman/obsah',
    author='The Foreman Project',
    author_email='foreman-dev@googlegroups.com',
    zip_safe=False,
    python_requires=">=3.9, <4",

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],

    keywords='ansible foreman packaging koji brew mock',

    packages=find_packages(exclude=['contrib', 'docs']),

    install_requires=[
        'ansible-core',
    ],

    extras_require={
        'argcomplete': ['argcomplete'],
    },

    package_data={
        'obsah': find_package_data('obsah', 'data'),
    },

    entry_points={
        'console_scripts': [
            'obsah=obsah:main',
        ],
    },

    project_urls={
        "Documentation": "https://obsah.readthedocs.io/en/latest/",
        "Source": "https://github.com/theforeman/obsah",
    },
)
